from typing import FrozenSet
import uuid
import json
import sys
import traceback
from decouple import config
from fastapi import FastAPI, HTTPException, Request
from fastapi.responses import JSONResponse
from starlette.middleware.base import BaseHTTPMiddleware
from decouple import config

# modules

from interfaces.interfaces_pd import Headers_PM, RequestContext_PM
from utils.jwt_validation import JwtValdationUtils
from logging.base_logger import APP_LOGGER
from context.vars import (
    headers_context,
    request_context,
    request_params_context,
    payload_context,
)


class HeaderValidationMiddleware(BaseHTTPMiddleware):
    def __init__(
        self,
        app: FastAPI,
        excluded_paths: FrozenSet[str] = frozenset(),
        authexpiryignore_paths: FrozenSet[str] = frozenset(),
    ):
        super().__init__(app)
        self.excluded_paths: FrozenSet[str] = frozenset(
            {"/docs", "/openapi.json", *excluded_paths}
        )

        self.authexpiryignore_paths: FrozenSet[str] = frozenset(authexpiryignore_paths)

    async def dispatch(self, request: Request, call_next):
        if any(request.url.path.startswith(path) for path in self.excluded_paths):
            return await call_next(request)
        headers = request.headers
        access_token = None
        verify_exp = True
        should_ignore_expiry = any(
            request.url.path.startswith(path) for path in self.authexpiryignore_paths
        )
        if "access_token" in request.cookies:
            access_token = request.cookies.get("access_token")
            verify_exp = not should_ignore_expiry
        elif "authorization" in headers:
            access_token = headers.get("authorization")
            verify_exp = not should_ignore_expiry
        access_token = request.cookies.get("access_token") or headers.get(
            "authorization"
        )
        try:
            if access_token:
                payload = JwtValdationUtils.validate_token(
                    access_token, verify_exp=verify_exp
                )
                headers_model = Headers_PM(
                    **dict(
                        correlationid=headers.get("correlationid", str(uuid.uuid4())),
                        username=payload["username"],
                        authorization=access_token,
                    )
                )
                headers_context.set(headers_model)
                return await call_next(request)

            if headers.get("x-api-key-2") == config(
                "X_API_KEY_PUBSUB-SERVICE_2"
            ) or headers.get("x-api-key-1") == config("X_API_KEY_PUBSUB-SERVICE_1"):
                headers_model = Headers_PM(
                    **dict(
                        correlationid=headers.get("correlationid", str(uuid.uuid4())),
                        username=headers.get("username", ""),
                        authorization=headers.get("authorization", ""),
                    )
                )
                headers_context.set(headers_model)
                return await call_next(request)

            return JSONResponse(
                {"detail": "Unauthorized: Missing or invalid credentials."},
                status_code=401,
            )

        except HTTPException as e:
            raise e
        except Exception as e:
            APP_LOGGER.error(f"Unexpected error in token validation: {str(e)}")
            raise HTTPException(status_code=500, detail="Internal server error")


class ExceptionMiddleware:
    def __init__(self, app):
        self.app = app

    def log_exception(self):
        exc_type, exc_value, exc_traceback = sys.exc_info()
        if exc_type is not None:
            traceback_info = traceback.extract_tb(exc_traceback)
            relevant_traceback = next(
                (
                    trace
                    for trace in reversed(traceback_info)
                    if str(config("SERVICENAME")) in trace.filename
                ),
                None,
            )
            if relevant_traceback:
                pathname, lineno, funcName, code_line = relevant_traceback
                error_data = {
                    "exception_type": str(exc_type.__name__),
                    "exception_value": str(exc_value),
                    "pathname": str(pathname),
                    "lineno": str(lineno),
                    "funcName": funcName,
                    "code_line": str(code_line),
                }
                APP_LOGGER.error(json.dumps(error_data))

    async def __call__(self, scope, receive, send):
        try:
            await self.app(scope, receive, send)
        except HTTPException as http_exception:
            self.log_exception()
            await self.handle_http_exception(http_exception, scope, receive, send)
        except Exception as e:
            self.log_exception()
            await self.handle_internal_server_error(e, scope, receive, send)

    async def handle_http_exception(self, http_exception, scope, receive, send):
        await http_exception(scope, receive, send)

    async def handle_internal_server_error(self, exception, scope, receive, send):
        error_message = dict(
            detail="Internal Server Error", error_message=str(exception)
        )
        response = JSONResponse(status_code=500, content=error_message)
        await response(scope, receive, send)


class ContextSetter:
    async def __call__(self, request: Request):
        if request.query_params:
            query_params = dict(request.query_params)
            request_params_context.set(query_params)
        try:
            body = await request.json()
            payload_context.set(body)
        except (json.JSONDecodeError, ValueError):
            pass

        request_path_method: RequestContext_PM = RequestContext_PM(
            request_trace_id=str(uuid.uuid4()),
            url_path=request.url.path,
            method=request.method,
        )
        request_context.set(request_path_method)



from .clients.http_client import HttpClient
from .clients.redis_client.sync_redisclient import SyncRedisClient
from .clients.redis_client.async_redisclient import AsyncRedisClient
from .clients.sqs_client.post_message import SyncSQSPusher
from .clients.sqs_client.polling import SQS_Manager
from .clients.sqs_client.initiator import SQSInitiator
from .logging.base_logger import APP_LOGGER
from .middlewares.message_middlewares import MessageMiddleware
from .middlewares.rest_middlewares import (
    HeaderValidationMiddleware,
    ExceptionMiddleware,
    ContextSetter,
)
from .ioc.singleton import SingletonMeta, Para_SingletonMeta
from .mongo_utils.transactions import MongoTransactionContext
from .mongo_utils.base_repository import BaseRepository
from .messaging_utils.async_task_handler import ProcessActionHandler



from typing import TypedDict, Required, Dict


class Jwk_TH(TypedDict, total=False):
    alg: Required[str]
    e: str
    kid: Required[str]
    kty: str
    n: str
    use: str


class Headers_TH(TypedDict):
    correlationid: str
    username: str
    authorization: str


class SSEMessage_TH(TypedDict):
    event: str
    data: Dict[str, str]


class SQSClientCallBackResponse_TH(TypedDict):
    allSuccess: bool
    correlationid: str



import asyncio
from redis.asyncio import Redis
from typing import AsyncGenerator, Optional
from interfaces.interfaces_th import SSEMessage_TH
from ioc.singleton import SingletonMeta
import json


class AsyncRedisClient(metaclass=SingletonMeta):
    def __init__(self, redis_connection: Optional[Redis] = None) -> None:
        if not hasattr(self, "redis") and isinstance(redis_connection, Redis):
            self.redis: Redis = redis_connection
        self.pubsub = self.redis.pubsub()

    async def publish(self, channel: str, message: SSEMessage_TH) -> int:
        return await self.redis.publish(channel=channel, message=json.dumps(message))

    async def subscribe(self, channel: str) -> None:
        await self.pubsub.subscribe(channel)

    async def unsubscribe(self, channel: str) -> None:
        await self.pubsub.unsubscribe(channel)

    async def get_message(self, timeout: float = 1.0) -> Optional[dict]:
        return await self.pubsub.get_message(timeout=timeout)

    async def listen(self, channel: str) -> AsyncGenerator[dict, None]:
        await self.subscribe(channel=channel)
        try:
            while True:
                message = await self.get_message(timeout=1.0)
                if message is not None:
                    yield message
                await asyncio.sleep(0.1)
        finally:
            await self.unsubscribe(channel=channel)

    async def close_connection(self) -> None:
        await self.redis.close()



import requests
from decouple import config
import uuid
from typing import Optional, Dict
from logging.base_logger import APP_LOGGER


class HttpClient:
    def __init__(
        self,
        headers: Dict[str, str],
        path: str,
        body=None,
        correlationid: Optional[str] = None,
        path_params: Optional[Dict[str, str]] = None,
        query_params: Optional[Dict[str, str]] = None,
        interServiceCall: bool = False,
        localhostCall: bool = False,
        portNumber: str = str(config("PORT")),
    ) -> None:
        self.interServiceCall = interServiceCall
        self.localhostCall = localhostCall
        self.__request_id = str(uuid.uuid4())
        self.__environment: str = str(config("ENVIRONMENT"))
        self.__domain: str = str(config("DOMAIN"))
        self.__portnumber: str = portNumber
        self.__headers = headers
        self.__path = path
        self.__body = body
        self.__path_params = path_params
        self.__query_params = query_params
        if correlationid is None:
            self.__correlationid = str(uuid.uuid4())
        else:
            self.__correlationid = correlationid
        self.__headers["correlationid"] = self.__correlationid

    def __construct_url(self) -> str:
        base_url = (
            f"https://{self.__environment}{self.__domain}/"
            if self.interServiceCall
            else f"http://localhost:{self.__portnumber}/"
        )
        url = base_url + self.__path
        if self.__path_params:
            url = url.format(**self.__path_params)
        return url

    def __send_request(self, method: str) -> requests.Response:
        url: str = self.__construct_url()
        APP_LOGGER.info(
            f"Sending {method} request to {url} with correlationId = {self.__correlationid} and requestid = {self.__request_id}"
        )
        try:
            response = requests.request(
                method=method,
                url=url,
                headers=self.__headers,
                params=self.__query_params,
                json=self.__body,
            )
            response.raise_for_status()  # Raise an exception for 4xx and 5xx status codes
            APP_LOGGER.info(
                f"Received response with status code {response.status_code} with correlationid = {self.__correlationid} and requestid  = {self.__request_id}"
            )
            return response
        except requests.exceptions.RequestException as e:
            APP_LOGGER.error(
                f"Request failed with correlationid = {self.__correlationid} and requestid = {self.__request_id}. Error: {str(e)}"
            )
            raise

    def get(self) -> requests.Response:
        return self.__send_request("GET")

    def post(self) -> requests.Response:
        return self.__send_request("POST")

    def patch(self) -> requests.Response:
        return self.__send_request("PATCH")

    def put(self) -> requests.Response:
        return self.__send_request("PUT")

    def delete(self) -> requests.Response:
        return self.__send_request("DELETE")
